package com.xefer.rebel;

import java.io.*;
import java.lang.*;
import java.util.*;

import javax.servlet.*;
import javax.servlet.http.*;

import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import org.w3c.dom.*;

/**
 * Encapsulates the XBEL document
 */
 
class XBELDocument extends XBELElement {

  /** The underlying XML Document that this class is managing */
  Document document;
  
  /**
   * constructor
   *
   * @param document Document the document that this class encapsulates
   */
  public XBELDocument(Document document) {

    this.document = document;
    this.element  = document.getDocumentElement();
    
  }

  /**
   * constructor
   *
   * @param element Element the document element of the resource
   */
   
  public XBELDocument(Element element) {
    super(element);  
    
    this.document = element.getOwnerDocument();
  }

  /**
   * constructor
   *
   * @param filename String the file to be parsed
   */
   
  public XBELDocument(String filename) throws ResourceException {
  
    InputStream istream = null;
    DocumentBuilderFactory factory;
    DocumentBuilder builder = null;
    
    try {
      
      factory = DocumentBuilderFactory.newInstance();
      factory.setExpandEntityReferences(false);
      builder = factory.newDocumentBuilder();
      
      File file = new File(filename);
      
      if(file.exists()) {

        istream = openStream(file);

        this.document = builder.parse(istream);
        this.element  = document.getDocumentElement();
        
      }
      else {
        
        // don't create the new File yet, don't want
        // an empty file hanging around if something fails
        this.document = builder.newDocument();
        this.element  = document.createElement("xbel");
        
        document.appendChild(element);
        
      }
    
    } catch (ParserConfigurationException ex) {
    
      throw new ResourceException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Can't build parser");
    
    } catch (org.xml.sax.SAXException ex) {
    
      throw new ResourceException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Can't parse local xbel document");
    
    } catch (IOException ex) {
    
      throw new ResourceException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Can't open local xbel document");
    
    } finally {
    
      closeStream(istream);
    
    }
  
  }
  
  /**
   * getElement retrieves the location at the given URL or throws a ResourceException
   * if not element can be located.
   *
   * @param path String representing the resources path.  
   * @return XBELElement at the given path
   */
  
  public XBELElement getElement(String path) throws ResourceException {
    
    return XBELElementFactory.newElement(document, path);
    
  }
  
  /**
   * getElement retrieves the element at the given URL path, generating a new
   * resource if the generate flag is true. For URL paths that end in a "/", the
   * generated element will be an XBELFolder (or XBELDocument if its the root), if
   * the URL does not end in a "/" the generated element will be a XBELBookmark
   *
   * @param path String of the path to the resource
   * @generate boolean true if the resource is to be generated fall if not
   * @return XBELElement generated or null if not generated and no resource found
   */
  
  public XBELElement getElement(String path, boolean generate) throws ResourceException {
    
    if (generate) {
      return XBELElementFactory.generateElement(document, path);
    } else {    
      return XBELElementFactory.getElement(document, path);
    } 
    
  }
  
  /**
   * put updates the resource element with the entity element
   *
   * @param resource XBELElement to be updated
   * @param entity XBELElement to be stored as the updated resource
   */
   
  public void putElement(XBELElement resource, XBELElement entity) throws ResourceException {
  
    resource.put(entity);
    
  }
  
  /**
   * deletes the element from the document
   *
   * @param resource XBELElement to be deleted
   */
   
  public void deleteElement(XBELElement resource) throws ResourceException {

    resource.delete();
  
  }
  
  /**
   * postElement
   *
   */
   
  public void postElement(XBELElement resource, XBELElement entity) throws ResourceException {
  
    resource.post(entity);
    
  }
  
  /**
   * delete
   *
   * Not allowed to delete the XBEL document itself
   */
   
  public void delete() throws ResourceException {

    throw new MethodNotAllowedException("DELETE not allowed; PUT an empty xbel to clear", "PUT,POST,GET,HEAD");
  
  }
  
  /**
   * Returns the resource Location URL based on the current url and the 
   * resource type
   *
   * @param url StringBuffer the URL location of the parent resource
   * @returns String
   */
   
  public String getLocationUrl(StringBuffer url) {
  
    String id = getId();
    
    return url.append(id).append("/").toString();
  
  }
  
  /**
   * Returns true if this XBELElement is directly locatable through a URL.  For
   * bookmark elements this is true.
   *
   * @returns boolean
   */
     
  public boolean isLocatable() {
    
    return true;
    
  }
  
  /**
   *
   * Write the given XML Document to the given file location
   *
   */
  
  public void write(ServletContext context, String filename) throws ResourceException {
  
    OutputStream ostream = null;
    
    try {
    
      InputStream istream =  context.getResourceAsStream("/WEB-INF/identity.xsl");
      
      TransformerFactory tFactory = TransformerFactory.newInstance();
      Transformer identity = tFactory.newTransformer(new StreamSource(istream));
  
      File output = new File(filename);
      if(output.exists() == false) {
        output.createNewFile();
      }
      ostream = new FileOutputStream(output.getCanonicalFile());

      Source source = new DOMSource(document);
      Result result = new StreamResult(ostream);

      identity.transform(source, result);
    
    } catch (Exception ex) {
    
      throw new ResourceException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, ex); 
    
    } finally {
    
      closeStream(ostream);
    
    }
        
  }

  /**
   *
   * Open the file and and return an input stream
   *
   * @param file File to be opened
   * @result FileInputStream an input stream to the passed-in File
   */

  protected FileInputStream openStream(File file) throws ResourceException {
  
    try {

      return new FileInputStream(file.getCanonicalFile());
      
    } catch (FileNotFoundException ex) {
    
      throw new ResourceException(HttpServletResponse.SC_NOT_FOUND, ex.getMessage());
    
    } catch (IOException ex) {
    
      throw new ResourceException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, ex.getMessage());
      
    } catch (SecurityException ex) {
    
      throw new ResourceException(HttpServletResponse.SC_FORBIDDEN, ex.getMessage());
    
    }
  
  }
  
  /**
   * silently closes the stream and absorbs and logs exceptions
   *
   * @param stream InputStream to be closed
   */
  
  protected void closeStream(InputStream stream) {
   
    if(stream != null) {
      try {
        stream.close();
      }
      catch (IOException ex ) {
        System.out.println("Error closing input stream");
      }
    }
  
  }
  
  /**
   * silently closes the stream and absorbs and logs exceptions
   *
   * @param stream OutputputStream to be closed
   */
  
  protected void closeStream(OutputStream stream) {
   
    if(stream != null) {
      try {
        stream.close();
      }
      catch (IOException ex ) {
        System.out.println("Error closing input stream");
      }

    }
  
  }

}